//--------------------------------------------------------------------
// name: Animal Farm
// desc: MUSIC 220B Hw3: Granular
//
// author: Cara Turnbull
// date: spring 2019
// 
// adapted from twilight-redux-kb.ck, written by Ge Wang (ge@ccrma.stanford.edu)
//
//
// keyboard/mouse control (also see kb() and mouseLoop() below):
//      Mouse X-axis: position of the grain in time (left is beginning, right is end)
//	    Mouse Y-axis: playrate of the grain (top is 0, bottom is 2)
//	    Left Arrow: make the grain length smaller
//	    Right Arrow: make the grain length larger
//	    Down Arrow: make the randomness value for grain position smaller
//	    Up Arrow: make the randomness value for grain position larger
//      keypad 1-6: select current animal (1=cow, 2=sheep, 3=horse, 4=obama, 5=bernie, 6=trump)
//      keypad +/-: connect or disconnect LiSa for current animal
//      <ENTER>: play clip* for current animal (if applicable)
//
//          *for the "trump" animal, there are multiple sound clips; the default is the
//              "trouble" clip, but you can select another clip by pressing a letter key
//              before pressing enter ('h'= "hombres", 'w'= "wall", 'b'= "bullshit",
//              't' = "trouble", and 's'= stupid)

//--------------------------------------------------------------------
// the instrument mode
Animal CURR_ANIMAL;
1 => int MODE_MOO;
2 => int MODE_BAA;
3 => int MODE_NEIGH;
4 => int MODE_OBAMA;
5 => int MODE_BERNIE;
6 => int MODE_TRUMP;
9 => int MODE_MAX;


// overall volume
.1 => float MAIN_VOLUME;
// grain duration base
10::ms => dur GRAIN_DURATION_BASE;
// grain duration (in multiples of duration base)
10 => float GRAIN_DURATION_FACTOR;
// grain duration base
GRAIN_DURATION_BASE * GRAIN_DURATION_FACTOR => dur GRAIN_LENGTH;
// factor relating grain duration to ramp up/down time
.5 => float GRAIN_RAMP_FACTOR;
// playback rate
1 => float GRAIN_PLAY_RATE;
// grain rate factor (0 means no change)
.05 => float GRAIN_RATE_FACTOR;
// grain position (0 start; 1 end)
0 => float GRAIN_POSITION;
// grain position randomization
.001 => float GRAIN_POSITION_RANDOM;
// grain jitter (0 == periodic fire rate)
1 => float GRAIN_FIRE_RANDOM;
// max lisa voices
30 => int LISA_MAX_VOICES;


// get sourcefiles
me.sourceDir() + "moo.wav" => string MOO;
me.sourceDir() + "baa.wav" => string BAA;
me.sourceDir() + "neigh.wav" => string NEIGH;

me.sourceDir() + "bad_hombres.wav" => string HOMBRES;
me.sourceDir() + "great_wall.wav" => string WALL;
me.sourceDir() + "political_bullshit.wav" => string BULLSHIT;
me.sourceDir() + "serious_trouble.wav" => string TROUBLE;
me.sourceDir() + "stupid_people.wav" => string STUPID;
me.sourceDir() + "core_values.wav" => string VALUES;
me.sourceDir() + "millions_of_people.wav" => string MILLIONS;


// load files into a LiSa (use one LiSa per sound)
load( MOO ) @=> LiSa @ lisaMoo;
load( BAA ) @=> LiSa @ lisaBaa;
load( NEIGH ) @=> LiSa @ lisaNeigh;
load( HOMBRES ) @=> LiSa @ lisaHombres;
load( WALL ) @=> LiSa @ lisaWall;
load( BULLSHIT ) @=> LiSa @ lisaBullshit;
load( TROUBLE ) @=> LiSa @ lisaTrouble;
load( STUPID ) @=> LiSa @ lisaStupid;
load( VALUES ) @=> LiSa @ lisaObama;
load( MILLIONS ) @=> LiSa @ lisaBernie;

LiSa lisaTrump;

// load files into a SndBufs (use one SndBuf per sound)
snd( MOO ) @=> SndBuf @ moo;
snd( BAA ) @=> SndBuf @ baa;
snd( NEIGH ) @=> SndBuf @ neigh;
snd( HOMBRES ) @=> SndBuf @ hombres;
snd( WALL ) @=> SndBuf @ wall;
snd( BULLSHIT ) @=> SndBuf @ bullshit;
snd( TROUBLE ) @=> SndBuf @ trouble;
snd( STUPID ) @=> SndBuf @ stupid;
snd( VALUES ) @=> SndBuf @ values;
snd( MILLIONS ) @=> SndBuf @ millions;


// patch it
PoleZero blocker => NRev reverb => Gain g => dac;
MAIN_VOLUME => g.gain;
.05 => reverb.mix;
.99 => blocker.blockZero; // pole location to block DC and ultra low frequencies



// reference to SndBuf to play any clips
SndBuf @ clip;

class Animal {
    string name;
    LiSa lisa;
    SndBuf source;
    int isPlaying;
    float grainPosition;
    float grainPlayRate;
    dur grainLength;
    float grainRandom;    
}

class Politician extends Animal {
    fun void loadLisa(LiSa clipLisa) {
        null @=> lisa;
        clipLisa @=> lisa;
    }
}

Animal cow;
"cow" => cow.name;
lisaMoo @=> cow.lisa;
moo @=> cow.source;
0 => cow.isPlaying;
GRAIN_POSITION => cow.grainPosition;
GRAIN_PLAY_RATE => cow.grainPlayRate;
GRAIN_LENGTH => cow.grainLength;
GRAIN_POSITION_RANDOM => cow.grainRandom;

Animal sheep;
"sheep" => sheep.name;
lisaBaa @=> sheep.lisa;
baa @=> sheep.source;
0 => sheep.isPlaying;
GRAIN_POSITION => sheep.grainPosition;
GRAIN_PLAY_RATE => sheep.grainPlayRate;
GRAIN_LENGTH => sheep.grainLength;
GRAIN_POSITION_RANDOM => sheep.grainRandom;

Animal horse;
"horse" => horse.name;
lisaNeigh @=> horse.lisa;
neigh @=> horse.source;
0 => horse.isPlaying;
GRAIN_POSITION => horse.grainPosition;
GRAIN_PLAY_RATE => horse.grainPlayRate;
GRAIN_LENGTH => horse.grainLength;
GRAIN_POSITION_RANDOM => horse.grainRandom;

Politician trump;
"trump" => trump.name;
lisaTrump @=> trump.lisa;
trouble @=> trump.source;
0 => trump.isPlaying;
GRAIN_POSITION => trump.grainPosition;
GRAIN_PLAY_RATE => trump.grainPlayRate;
GRAIN_LENGTH => trump.grainLength;
GRAIN_POSITION_RANDOM => trump.grainRandom;

Politician obama;
"obama" => obama.name;
lisaObama @=> obama.lisa;
values @=> obama.source;
0 => obama.isPlaying;
GRAIN_POSITION => obama.grainPosition;
GRAIN_PLAY_RATE => obama.grainPlayRate;
GRAIN_LENGTH => obama.grainLength;
GRAIN_POSITION_RANDOM => obama.grainRandom;

Politician bernie;
"bernie" => bernie.name;
lisaBernie @=> bernie.lisa;
millions @=> bernie.source;
0 => bernie.isPlaying;
GRAIN_POSITION => bernie.grainPosition;
GRAIN_PLAY_RATE => bernie.grainPlayRate;
GRAIN_LENGTH => bernie.grainLength;
GRAIN_POSITION_RANDOM => bernie.grainRandom;

// select mode
fun void setAnimal( int which )
{
    if( clip != null) clip =< reverb;
    null @=> clip;
    // check mode
    if( which == MODE_MOO ) {
        cow @=> CURR_ANIMAL;
        cow.source @=> clip;
    }
    else if( which == MODE_BAA ) {
        sheep @=> CURR_ANIMAL;
        sheep.source @=> clip;
    }
    else if( which == MODE_NEIGH ) {
        horse @=> CURR_ANIMAL;
        horse.source @=> clip;
    }
    else if( which == MODE_OBAMA ) {
        obama @=> CURR_ANIMAL;
        obama.loadLisa(obama.lisa);
        obama.source @=> clip;
    }
    else if( which == MODE_BERNIE ) {
        bernie @=> CURR_ANIMAL;
        bernie.loadLisa(bernie.lisa);
        bernie.source @=> clip;
    }
    else if( which == MODE_TRUMP ) {
        trump @=> CURR_ANIMAL;
        trump.loadLisa(lisaHombres);
        trump.source @=> clip;
    }


}


// HID objects
Hid hi;
HidMsg msg;

// which keyboard
0 => int device;
// get from command line
if( me.args() ) me.arg(0) => Std.atoi => device;

// open keyboard 0, exit on fail
if( !hi.openKeyboard( device ) ) me.exit();
// log
<<< "keyboard '" + hi.name() + "' ready", "" >>>;

// keys
45 => int KEY_MINUS;
43 => int KEY_PLUS;
54 => int KEY_COMMA;
55 => int KEY_PERIOD;
79 => int KEY_RIGHT;
80 => int KEY_LEFT;
81 => int KEY_DOWN;
82 => int KEY_UP;
72 => int KEY_H;
87 => int KEY_W;
66 => int KEY_B;
84 => int KEY_T;
83 => int KEY_S;
10 => int KEY_ENTER;

// spork it
spork ~ print();
spork ~ kb();
spork ~ mouseLoop();

// main loop

while( true )
{
    // fire a grain
    fireGrain(cow);
    fireGrain(sheep);
    fireGrain(horse);
    fireGrain(obama);
    fireGrain(bernie);
    fireGrain(trump);
    // amount here naturally controls amount of overlap between grains
    GRAIN_LENGTH / 8 + Math.random2f(0,GRAIN_FIRE_RANDOM)::ms => now;
}

// fire!
fun void fireGrain(Animal animal)
{
    // grain length
    animal.grainLength => dur grainLen;
    // ramp time
    animal.grainLength * GRAIN_RAMP_FACTOR => dur rampTime;
    // play pos
    animal.grainPosition + Math.random2f(0,GRAIN_POSITION_RANDOM) => float pos;
    // a grain
    if( animal.lisa != null && pos >= 0 )
        spork ~ grain( animal.lisa, pos * animal.lisa.duration(), grainLen, rampTime, rampTime, 
        animal.grainPlayRate );
   
}

// grain sporkee
fun void grain( LiSa @ lisa, dur pos, dur grainLen, dur rampUp, dur rampDown, float rate )
{
    // get a voice to use
    lisa.getVoice() => int voice;

    // if available
    if( voice > -1 )
    {
        // set rate
        lisa.rate( voice, rate );
        // set playhead
        lisa.playPos( voice, pos );
        // ramp up
        lisa.rampUp( voice, rampUp );
        // wait
        (grainLen - rampUp) => now;
        // ramp down
        lisa.rampDown( voice, rampDown );
        // wait
        rampDown => now;
    }
}

// print
fun void print()
{
    // time loop
    while( true )
    {
        // values
        <<< "animal", CURR_ANIMAL.name, "pos:", CURR_ANIMAL.grainPosition, "random:", CURR_ANIMAL.grainRandom,
            "rate:", CURR_ANIMAL.grainPlayRate, "size:", CURR_ANIMAL.grainLength/second >>>;
        // advance time
        100::ms => now;
    }
}

// keyboard
fun void kb()
{
    // infinite event loop
    while( true )
    {
        // wait on HidIn as event
        hi => now;
        
        // messages received
        while( hi.recv( msg ) )
        {
            // button down
            if( msg.isButtonDown() )
            {

                if( msg.key == KEY_LEFT )
                {
                    .95 *=> CURR_ANIMAL.grainLength;
                    if( CURR_ANIMAL.grainLength < 1::ms ) 1::ms => CURR_ANIMAL.grainLength;
                }
                else if( msg.key == KEY_RIGHT )
                {
                    1.305 *=> CURR_ANIMAL.grainLength;
                    if( CURR_ANIMAL.grainLength > 1::second ) 1::second => CURR_ANIMAL.grainLength;
                }
                else if( msg.key == KEY_DOWN )
                {
                    .9 *=> CURR_ANIMAL.grainRandom;
                    if( CURR_ANIMAL.grainRandom < .000001 ) .000001 => CURR_ANIMAL.grainRandom;
                }
                else if( msg.key == KEY_UP )
                {
                    1.1 *=> CURR_ANIMAL.grainRandom;
                    if( CURR_ANIMAL.grainRandom > 1 ) 1 => CURR_ANIMAL.grainRandom;
                }
                else if ( msg.ascii >= 48 && msg.ascii <= 54) {
                    msg.ascii - 48 => int m;
                    setAnimal( m );
                }
                else if (msg.ascii == KEY_MINUS) {
                    if(CURR_ANIMAL.isPlaying == 1) {
                        CURR_ANIMAL.lisa.chan(0) =< blocker;
                        0 => CURR_ANIMAL.isPlaying;
                    }
                }
                else if (msg.ascii == KEY_PLUS) {
                    if(CURR_ANIMAL.isPlaying == 0) {
                        CURR_ANIMAL.lisa.chan(0) => blocker;
                        1 => CURR_ANIMAL.isPlaying; 
                    }
                }
                else if ( msg.ascii == KEY_H ) {
                    if(CURR_ANIMAL.name == "trump") {
                        CURR_ANIMAL.lisa.chan(0) =< blocker;
                        trump.loadLisa(lisaHombres);
                        CURR_ANIMAL.lisa.chan(0) => blocker;
                        hombres @=> clip;
                    }
                }
                else if ( msg.ascii == KEY_W ) {
                    if(CURR_ANIMAL.name == "trump") {
                        CURR_ANIMAL.lisa.chan(0) =< blocker;
                        trump.loadLisa(lisaWall);
                        CURR_ANIMAL.lisa.chan(0) => blocker;
                        wall @=> clip;
                    }
                }
                else if ( msg.ascii == KEY_B ) {
                    if(CURR_ANIMAL.name == "trump") {
                        CURR_ANIMAL.lisa.chan(0) =< blocker;
                        trump.loadLisa(lisaBullshit);
                        CURR_ANIMAL.lisa.chan(0) => blocker;
                        bullshit @=> clip;
                    }
                }
                else if ( msg.ascii == KEY_T ) {
                    if(CURR_ANIMAL.name == "trump") {
                        CURR_ANIMAL.lisa.chan(0) =< blocker;
                        trump.loadLisa(lisaTrouble);
                        CURR_ANIMAL.lisa.chan(0) => blocker;
                        trouble @=> clip;
                    }
                }
                else if ( msg.ascii == KEY_S ) {
                    if(CURR_ANIMAL.name == "trump") {
                        CURR_ANIMAL.lisa.chan(0) =< blocker;
                        trump.loadLisa(lisaStupid);
                        CURR_ANIMAL.lisa.chan(0) => blocker;
                        stupid @=> clip;
                    }
                }
                else if ( msg.ascii == KEY_ENTER ) {
                    clip.pos(0);
                    clip.rate(1);
                    if( clip != null ) clip => reverb;
                }      
            }
        }
    }
}

fun void mouseLoop() {
    Hid hi;
    HidMsg msg;
    
    float xPos;
    float yPos;
    // which mouse
    0 => int device;
    // get from command line
    if( me.args() ) me.arg(0) => Std.atoi => device;
    // open mouse 0, exit on fail
    if( !hi.openMouse( device ) ) me.exit();
    //<<< "mouse '" + hi.name() + "' ready", "" >>>;

    while( true) {
        hi => now;
        
        while( hi.recv( msg ) ) {
           if ( msg.isMouseMotion()){ // mouse motion--clip values between -1 and 1
               if (Math.fabs(msg.deltaX * .001 + xPos) < 1) {
                   msg.deltaX * .001 + xPos => xPos;
               }
               
               if ( Math.fabs(msg.deltaY * .001 + yPos) < 1) {
                   msg.deltaY * .001 + yPos => yPos;
               }               
               //convert values from [-1 1] to [0 1]
               (xPos + 1) / 2 => float newXPos;
               (yPos + 1) / 2 => float newYPos;
               newXPos => CURR_ANIMAL.grainPosition;
               1.25 * newYPos => CURR_ANIMAL.grainPlayRate;
                           
           } 
           if( msg.isButtonDown() ) {
              
           } 
           if( msg.isButtonUp() ) {

           }
        }
    }
}

// load file into a LiSa
fun LiSa load( string filename )
{
    // sound buffer
    SndBuf buffy;
    // load it
    filename => buffy.read;
    
    // new LiSa
    LiSa lisa;
    // set duration
    buffy.samples()::samp => lisa.duration;
    
    // transfer values from SndBuf to LiSa
    for ( 0 => int j; j < buffy.samples(); j++ ) {
    
        (buffy.valueAt(j * buffy.channels()), j::samp) => lisa.valueAt;
    
    }
    
    // set LiSa parameters
    lisa.play( false );
    lisa.loop( false );
    lisa.maxVoices( LISA_MAX_VOICES );
    
    return lisa;
}

// load file into a SndBuf
fun SndBuf snd( string filename )
{
    // sound buffer
    SndBuf buffy;
    // load it
    filename => buffy.read;
    
    // set position to end (so it doesn't play)
    buffy.samples() => buffy.pos;
    
    return buffy;
}